//
// Functions for mathematical intersection tests
//
// by Jonathan Kreuzer
// http://www.3dkingdoms.com/weekly
//
#include <math.h>
#include "vec3.h"
#include "mesh.h"
#include "intersection.h"

//
// returns true if point P is in the 4-plane frustum
//
bool PointInFrustum( const CVec3 &P, CVec3 Normals[4], CVec3 Points[8] )
{
	for ( int x = 0; x < 4; x++ )
	{
		if ( Normals[x].Dot( P - Points[x*2] ) > 0 ) return false;
	}
    return true;
}

//
// LineInFrustum is much slower than it could be. I don't think it will be called very often, if it ever is I'll rewrite it.
// It just constructs a 4-plane frustum box out of triangles, and uses the Line-Triangle test.
//
bool LineInFrustum( const CVec3 &LP1, const CVec3 &LP2, CVec3 Points[8] )
{
	CVec3 HitP;
	int List[24] = { 0, 1, 2, 1, 2, 3, 0, 1, 6, 1, 6, 7, 6, 7, 5, 6, 5, 4, 2, 3, 5, 2, 5, 4 };
	
	for ( int x = 0; x < 8; x++ ) 
	{
		if ( CheckLineTri( LP1, LP2, Points[ List[x*3] ], Points[ List[x*3+1] ], Points[ List[x*3+2] ], HitP ) )
			return true;
	}
			
	return false;
}

//
// returns false if the triangle is not within the frustum
//
bool TriInFrustum( CVec3 TriP[3], CVec3 Normals[4], CVec3 Points[8] )
{
 int i;
 // If all 3 points are to one side any frustum plane return false
 for( int plane = 0; plane < 4; plane++ )
	{
	for ( i = 0; i < 3; i++ )
		{
		if ( Normals[plane].Dot( TriP[i] - Points[plane*2] ) < 0 ) break;
		}
	if ( i >= 3 ) return false;
	}
  // If any point is in the frustum, return true
  for ( i = 0; i < 3; i++ )
	{
	if ( PointInFrustum( TriP[i], Normals, Points ) ) return true;
	}
  // If we didn't get quick result, do a slower but accurate test.
  // If any of the lines of the triangle are in the frustum, the triangle is in the frustum
  if ( LineInFrustum( TriP[0], TriP[1], Points ) ) return true;
  if ( LineInFrustum( TriP[1], TriP[2], Points ) ) return true;
  if ( LineInFrustum( TriP[2], TriP[0], Points ) ) return true;
  
  // If the frustum is completely inside the triangle, any frustum line into the screen will intersect the triangle
  CVec3 HitP;
  if ( CheckLineTri( Points[0], Points[1], TriP[0], TriP[1], TriP[2], HitP ) ) return true;
  
return false;
}

//
// CheckLineBox Helper Functions
//
bool inline GetIntersection( float fDst1, float fDst2, CVec3 P1, CVec3 P2, CVec3 &Hit) {
if ( (fDst1 * fDst2) >= 0.0f) return false;
if ( fDst1 == fDst2) return false;
Hit = P1 + (P2-P1) * ( -fDst1/(fDst2-fDst1) );
return true;
}

bool inline InBox( CVec3 Hit, CVec3 B1, CVec3 B2, const int Axis ) {
if ( Axis == 1 && Hit.z > B1.z && Hit.z < B2.z && Hit.y > B1.y && Hit.y < B2.y) return true;
if ( Axis == 2 && Hit.z > B1.z && Hit.z < B2.z && Hit.x > B1.x && Hit.x < B2.x) return true;
if ( Axis == 3 && Hit.x > B1.x && Hit.x < B2.x && Hit.y > B1.y && Hit.y < B2.y) return true;
return false;
}

//  
//  returns true if line (L1, L2) intersects with axis-aligned box (B1, B2) 
//  The point of intersection is returned in HitP
//
bool CheckLineBox( CVec3 L1, CVec3 L2, CVec3 B1, CVec3 B2, CVec3 &HitP )
{
// Check for a quick exit if the line is completely to one side of the box
if (L2.x < B1.x && L1.x < B1.x) return false;
if (L2.x > B2.x && L1.x > B2.x) return false;
if (L2.y < B1.y && L1.y < B1.y) return false;
if (L2.y > B2.y && L1.y > B2.y) return false;
if (L2.z < B1.z && L1.z < B1.z) return false;
if (L2.z > B2.z && L1.z > B2.z) return false;

// Check if the line originates in the box
if (L1.x > B1.x && L1.x < B2.x && L1.y > B1.y && L1.y < B2.y && L1.z > B1.z && L1.z < B2.z ) {
	HitP = L1; 
	return true;
	}
	
// Check for a line intersection with each side of the box
if ( GetIntersection( L1.x-B1.x, L2.x-B1.x, L1, L2, HitP) && InBox( HitP, B1, B2, 1 ) ) return true;
if ( GetIntersection( L1.y-B1.y, L2.y-B1.y, L1, L2, HitP) && InBox( HitP, B1, B2, 2 ) ) return true;
if ( GetIntersection( L1.z-B1.z, L2.z-B1.z, L1, L2, HitP) && InBox( HitP, B1, B2, 3 ) ) return true;
if ( GetIntersection( L1.x-B2.x, L2.x-B2.x, L1, L2, HitP) && InBox( HitP, B1, B2, 1 ) ) return true;
if ( GetIntersection( L1.y-B2.y, L2.y-B2.y, L1, L2, HitP) && InBox( HitP, B1, B2, 2 ) ) return true;
if ( GetIntersection( L1.z-B2.z, L2.z-B2.z, L1, L2, HitP) && InBox( HitP, B1, B2, 3 ) ) return true;

return false;
}

//
//  returns true if line (L1, L2) intersects with triangle( PV1, PV2, PV3 )
//  The point of intersection is returned in HitP
//
bool CheckLineTri( const CVec3 &L1, const CVec3 &L2, 
				const CVec3 &PV1, const CVec3 &PV2, const CVec3 &PV3, 
				CVec3 &HitP )
{
// Find Triangle Normal, would be quicker to have these computed already
CVec3 Normal = ( PV2 - PV1 ).Cross( PV3 - PV1 );
Normal.Normalize();

// Find distance from L1 and L2 to the plane defined by the triangle
float fDst1 = (L1-PV1).Dot( Normal );
float fDst2 = (L2-PV1).Dot( Normal );

if ( (fDst1 * fDst2) >= 0.0f) return false;  // line doesn't cross the triangle.
if ( fDst1 == fDst2) return false; // line and plane are parallel

// Find point on the line that intersects with the plane
CVec3 IntersectP = L1 + (L2-L1) * ( -fDst1/(fDst2-fDst1) );

// Find if the interesection point lies inside the triangle by testing it against all edges
// would be quicker to have these edge planes precomputed too
CVec3 vTest;
vTest = Normal.Cross( PV2-PV1 );
if ( vTest.Dot( IntersectP-PV1 ) < 0.0f ) return false;
vTest = Normal.Cross( PV3-PV2 );
if ( vTest.Dot( IntersectP-PV2 ) < 0.0f ) return false;
vTest = Normal.Cross( PV1-PV3 );
if ( vTest.Dot( IntersectP-PV1 ) < 0.0f ) return false;

HitP = IntersectP;

return true;
}

//
// Ray Triangle Test ( Moller-Haines? )
// returns true if ray ( Ro, Rdir ) intersects with triangle( PV1, PV2, PV3 )
// on intersection returns baycentric coordinates in du and dv
// note: Intersection Point = PV1 + (PV2-PV1) * du + (PV3-PV1) * dv
//
bool CheckRayTri( const CVec3& Ro, const CVec3& Rdir,
			const CVec3& PV1, const CVec3& PV2, const CVec3& PV3, 
			float &du, float &dv )
{
	CVec3 Edge1 = PV2 - PV1;
	CVec3 Edge2 = PV3 - PV1;

	// Calculate the determinant
	CVec3 pvec = Rdir.Cross( Edge2 );
	float det = fabsf( Edge1.Dot( pvec ) );

	// The ray is not on the triangle if u < 0 or v < 0 or u + v > det
	CVec3 tvec = Ro - PV1;

	float u = tvec.Dot( pvec );
	if ( u < 0 || u > det )
		return false;

	float v = Rdir.Dot( tvec.Cross( Edge1 ) );
	if ( v < 0 || u + v > det )
		return false;

	// On Intersection, return barycentric coordinates
	du = u/det;
	dv = v/det;

	return true;
}