//
// Mesh Class
// 
// Right now just loads simple meshes with only verts and triangles
// Note: Each triangle is defined by the indices of 3 vertices
//
#define _CRT_SECURE_NO_DEPRECATE
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <math.h>
#include <stdio.h>

#include "vec3.h"
#include "matrix.h"
#include "mesh.h"
#include "renderInterface.h"
#include "intersection.h"
#include "datatransfer.h"

CMesh::CMesh()
{
	m_nbVerts = 0;
	m_nbTris = 0;

	m_pVerts = NULL;
	m_pUVs = NULL;
	m_pVertNormals = NULL;
	m_pTris = NULL;
	m_pTriFlags = NULL;
	m_pTriNormals = NULL;

	m_SelectMode = SELECT_ADD;
}

CMesh::~CMesh()
{
	Deallocate();
}

void CMesh::Allocate( int nbVerts, int nbTris )
{
	m_pVerts = new CVec3[ nbVerts + 1 ];
	m_pUVs  = new CVec2[ nbVerts + 1 ];
	m_pVertNormals = new CVec3[ nbVerts + 1 ];
	m_pTris = new unsigned int[ nbTris*3 + 1 ];
	m_pTriFlags = new int[ nbTris + 1 ];
	m_pTriNormals = new CVec3[ nbTris + 1 ];
}

void CMesh::Deallocate()
{
	delete[] m_pVerts;
	delete[] m_pUVs;
	delete[] m_pVertNormals;
	delete[] m_pTris;
	delete[] m_pTriFlags;
	delete[] m_pTriNormals;
}

//
// Loads a mesh file from disk
//
bool CMesh::Load( const char *sFilename )
{
	FILE *pFile = fopen( sFilename, "rb" );

	if ( pFile ) 
		{
		CDataTransfer LoadIO( pFile, CDataTransfer::FILE_LOAD );
		// Read the File
		if ( DataIO( LoadIO ) )
			{
			ClearTriFlags();
			CalculateFaceNormals();
			CalculateVertNormals();

			fclose( pFile );
			return true;
			}
		
		fclose( pFile );
		return false;
		}
	return false;
}

//
// DataIO is used to read or write from/to a file
// ( in the future this could also be used for cloning, stats, etc. )
//
bool CMesh::DataIO( CDataTransfer &IO )
{
	char Verify[16] = "TestMesh";
	int Version = 1;

	// First read file type verification string
	IO.Data( Verify, 8 );
	if ( IO.IsRead() && memcmp( Verify, "TestMesh", 8 ) != 0 ) return false;

	// Then read the version
	IO << Version;

	// Then the mesh data
	IO << m_nbVerts;
	IO << m_nbTris;
	
	if ( IO.IsRead() ) {
		Deallocate(); // Just in case we're reloading into the same mesh
		Allocate( m_nbVerts, m_nbTris );
	}

	IO.Data( m_pVerts, sizeof(CVec3) * m_nbVerts );
	IO.Data( m_pUVs, sizeof(float) * 2 * m_nbVerts );
	IO.Data( m_pTris , 3 * sizeof(int) * m_nbTris );
	return true;
}

//
// Compute the face normals using the crossproduct of the edges and put them in m_pTriNormals
//
void CMesh::CalculateFaceNormals( )
{
	for ( int nTri = 0; nTri < m_nbTris; nTri++ )
		{
		int nV = nTri*3;
		CVec3 P0 = m_pVerts[ m_pTris[ nV ] ];
		CVec3 P1 = m_pVerts[ m_pTris[ nV+1 ] ];
		CVec3 P2 = m_pVerts[ m_pTris[ nV+2 ] ];
		m_pTriNormals[ nTri ] = (P1-P0).Cross(P2-P0);
		m_pTriNormals[ nTri ].Normalize();
		}
}

//
// Compute normals for each vertex by averaging the normals of adjacent faces.
// Requires face normals to be computed first.
//
void CMesh::CalculateVertNormals( )
{
	int i;
	for ( i = 0; i < m_nbVerts; i++ )
	{
		m_pVertNormals[ i ].SetZero();
	}
	for ( int nTri = 0; nTri < m_nbTris; nTri++ )
	{
		int Index = nTri*3;
		m_pVertNormals[ m_pTris[Index] ]   += m_pTriNormals[ nTri ];
		m_pVertNormals[ m_pTris[Index+1] ] += m_pTriNormals[ nTri ];
		m_pVertNormals[ m_pTris[Index+2] ] += m_pTriNormals[ nTri ];
	}
	for ( i = 0; i < m_nbVerts; i++ )
	{
		m_pVertNormals[ i ].Normalize();
	}
}

// Zero out m_pTriFlags
void CMesh::ClearTriFlags()
{
	memset( m_pTriFlags, 0, m_nbTris * sizeof(int) );
}

//
// Rendering
//
void CMesh::Render( CRenderer *pRender, bool bWireframe )
{
	// Draw the solid texture mapped mesh
	if ( bWireframe ) pRender->SetLineMode( true );
	pRender->SetColor( 1.0f, 1.0f, 1.0f );
	pRender->DrawMeshTextured( m_pVerts, m_pVertNormals, m_pTris, (float *)m_pUVs, m_nbTris );
	if ( bWireframe ) pRender->SetLineMode( false );

	// Highlight Selected Triangles
	for (int nTri = 0; nTri < m_nbTris; nTri++ )
		{
		if ( (m_pTriFlags[ nTri ] & TF_SELECTED) ) {
			pRender->HighlightTri( m_pVerts[ m_pTris[nTri*3] ], m_pVerts[ m_pTris[nTri*3+1] ], m_pVerts[ m_pTris[nTri*3+2 ]] );
			}
		}
}

//
// Render the vertex normals as blue lines
//
void CMesh::RenderNormals( CRenderer *pRender, float NormalLength )
{
	pRender->SetColor( 0.0f, .5f, 1.0f );
	for ( int i = 0; i < m_nbVerts; i++ )
	{
		CVec3 P = m_pVerts[ i ];
		pRender->DrawLine( P, P + m_pVertNormals[i] * NormalLength );
	}
}

//
// Render the vertices as red dots
//
void CMesh::RenderVertices( CRenderer *pRender )
{
	pRender->SetColor( 1.0f, 0.2f, 0.0f );
	pRender->SetPointSize( 2.5f );
	for ( int i = 0; i < m_nbVerts; i++ )
	{
		pRender->DrawPoint( m_pVerts[ i ] );
	}
}

//
// Render the UVs as a wireframe
//
void CMesh::RenderUVs( CRenderer *pRender )
{
	for ( int nTri = 0; nTri < m_nbTris; nTri++ )
	{
		if ( (m_pTriFlags[ nTri ] & TF_SELECTED) ) {
			pRender->SetColor( 1.0f, 1.0f, 1.0f );
		} else {
			pRender->SetColor( .5f, .5f, .5f );
		}

		pRender->DrawTri2D( m_pUVs[ m_pTris[nTri*3] ], m_pUVs[ m_pTris[nTri*3+1] ], m_pUVs[ m_pTris[nTri*3+2 ]] );
	}
	pRender->SetColor( 1.0f, 1.0f, 1.0f );
	pRender->DrawTexturedRect( 0.0f, 0.0f, 1.0f, 1.0f );
}

//
// MESH SELECTION
//
void CMesh::ClearSelection( )
{
	for (int nTri = 0; nTri < m_nbTris; nTri++ )
	{
		m_pTriFlags[ nTri ] &= ~TF_SELECTED;
	}
}

void CMesh::SetSelectionMode( eSelectionMode Mode )
{
	m_SelectMode = Mode;
}

// Sets the TF_SELECTED flag for the triangle at nTri
void CMesh::SelectTriangle( int nTri ) 
{
	if ( nTri < 0 || nTri >= m_nbTris ) return;
	if ( m_SelectMode == SELECT_ADD ) m_pTriFlags[ nTri ] |= TF_SELECTED;
	if ( m_SelectMode == SELECT_SUB ) m_pTriFlags[ nTri ] &= ~TF_SELECTED;
}

// Sets the TF_BACKFACING flag for triangles that are backfacing from CamPos's point of view
void CMesh::MarkBackfacing( CVec3 CamPos )
{
	for (int nTri = 0; nTri < m_nbTris; nTri++ )
	{
		if ( m_pTriNormals[ nTri ].Dot( CamPos - m_pVerts[ m_pTris[ nTri*3 ] ] ) < 0 ) 
			m_pTriFlags[ nTri ] |= TF_BACKFACING;
		else
			m_pTriFlags[ nTri ] &= ~TF_BACKFACING;
	}
}

// Select the closest triangle the line( LP1, LP2 ) intersects with
bool CMesh::LineSelect( const CVec3 &LP1, const CVec3 &LP2 )
{
	TriIntersect HitTri = LineIntersection( LP1, LP2 );
	if ( HitTri.Index >= 0 ) {
		SelectTriangle( HitTri.Index );
		return true;
	}
	return false;
}

//
// Test a line( LP1-LP2 ) against the mesh
// returns closest front-facing triangle
//
TriIntersect CMesh::LineIntersection( const CVec3 &LP1, const CVec3 &LP2 )
{
	float fDistance = 1000000000.0f;
	TriIntersect HitTri;
	HitTri.Index = -1;

	CVec3 Dir = LP2-LP1;
	Dir.Normalize();
	
	for (int nTri = 0; nTri < m_nbTris; nTri++ )
		{
		if ( m_pTriFlags[ nTri ] & TF_BACKFACING ) continue; // Check only front facing triangles
		
		int nV = nTri*3;
		float du, dv;
		CVec3 V1 = m_pVerts[ m_pTris[nV] ];
		CVec3 V2 = m_pVerts[ m_pTris[nV+1] ];
		CVec3 V3 = m_pVerts[ m_pTris[nV+2] ];
		bool bHit = CheckRayTri( LP1, Dir, V1, V2, V3, du, dv );

		if ( bHit ) 
			{
			CVec3 HitP = V1 + (V2-V1) * du + (V3-V1) * dv;

			if ( HitP.Distance( LP1 ) < fDistance ) 
				{
				m_HitP = HitP;
				fDistance = HitP.Distance( LP1 );

				HitTri.Index = nTri;
				HitTri.deltaEdge1 = du;
				HitTri.deltaEdge2 = dv;
				}
			}
		}
	return HitTri;
}

//
// Returns the Texture UV coordinates of triangle( nTri ) at barycentric coordinates (Edge1, Edge2)
//
CVec2 CMesh::GetTexCoords( const TriIntersect& Tri )
{
	CVec2 T1 = m_pUVs[ m_pTris[Tri.Index*3] ];
	CVec2 T2 = m_pUVs[ m_pTris[Tri.Index*3+1] ];
	CVec2 T3 = m_pUVs[ m_pTris[Tri.Index*3+2] ];
	CVec2 Coord;
	Coord.x = T1.x + (T2.x-T1.x) * Tri.deltaEdge1 + (T3.x-T1.x) * Tri.deltaEdge2;
	Coord.y = T1.y + (T2.y-T1.y) * Tri.deltaEdge1 + (T3.y-T1.y) * Tri.deltaEdge2;
	return Coord;
}

//
// Select mesh triangles that are in a frustum defined by 8 Points( and 4 face normals )
// I'm currently ignoring near and far planes, but they could be easily added
//
int CMesh::FrustumSelect( CVec3 Normals[4], CVec3 Points[8] )
{
	int nbHits = 0;
	CVec3 Tri[3];
	
	for (int nTri = 0; nTri < m_nbTris; nTri++ )
		{	
		int nV = nTri*3;
		Tri[0] = m_pVerts[ m_pTris[ nV ] ];
		Tri[1] = m_pVerts[ m_pTris[ nV+1 ] ];
		Tri[2] = m_pVerts[ m_pTris[ nV+2 ] ];

		if ( TriInFrustum( Tri, Normals, Points ) )
			{
			SelectTriangle( nTri );			
			nbHits++;			
			}
		}
	
	return nbHits;
}

//
// Get the triangle the line (LP1, LP2) intersects with.
// If it does intersect, return the texture coordinates of the intersection
//
bool CMesh::GetTexCoordsFromLine( const CVec3 &LP1, const CVec3 &LP2, CVec2& TexCoords )
{
	TriIntersect HitTri = LineIntersection( LP1, LP2 );
	if ( HitTri.Index >= 0 ) 
	{
		TexCoords = GetTexCoords( HitTri );
		return true;
	}
	return false;
}