//
// A Rendering class
// All rendering is done through this class. 
//
// There's a couple of reasons for this.
// 1. Resuse application specific rendering functions
// 2. To make it easier to replace OpenGL with another renderer, eg. Direct3D.
//

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <gl/gl.h>
#include <gl/glu.h>
#include <math.h>
#include "vec3.h"
#include "matrix.h"
#include "mesh.h"
#include "paint.h"
#include "renderinterface.h"

// Ctor 
CRenderer::CRenderer()
{
	m_hWnd = NULL;
	m_hDC = NULL;
	m_hGLRC = NULL;
	m_bInitialized = false;
	m_bLighting = false;
	m_bTextured = false;
	m_BindTexId = 0;
	m_nbTextures = 0;
}

int CRenderer::IsInitialized( )
{
	return m_bInitialized;
}

//
// Initialization
//
int CRenderer::InitRenderWindow( HWND hWnd, int bStencil, int bDoubleBuffer )
{
	PIXELFORMATDESCRIPTOR pfd;
	int format;
	
	m_hWnd = hWnd;
	m_hDC = GetDC( m_hWnd );

	// set the pixel format for the DC
	ZeroMemory( &pfd, sizeof(  PIXELFORMATDESCRIPTOR ) );
	pfd.nSize = sizeof(  PIXELFORMATDESCRIPTOR );
	pfd.nVersion = 1;
	pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL;
	if ( bDoubleBuffer ) pfd.dwFlags |= PFD_DOUBLEBUFFER;
	pfd.iPixelType = PFD_TYPE_RGBA;
	pfd.cColorBits = 24;
	pfd.cDepthBits = 16;
	if ( bStencil ) pfd.cStencilBits = 8; 
	pfd.iLayerType = PFD_MAIN_PLANE;

	format = ChoosePixelFormat( m_hDC, &pfd );
    if ( !SetPixelFormat( m_hDC, format, &pfd ) )
	{
		return false;
	}

	// create the context
	m_hGLRC = wglCreateContext( m_hDC );
	
	if ( wglMakeCurrent( m_hDC, m_hGLRC ) ) 
	{
		m_bInitialized = true;
		SetInitialState();
	}
	
	return true;
}

//
//
void CRenderer::Shutdown( )
{
	wglMakeCurrent( NULL, NULL );
	wglDeleteContext( m_hGLRC );
	ReleaseDC( m_hWnd, m_hDC );
}

// 
//
void CRenderer::ClearBuffer( )
{
	glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
}

// 
//
void CRenderer::BufferSwap( )
{
	glFlush();
	SwapBuffers( m_hDC );
}

//
// Setting States
//
void CRenderer::SetInitialState( )
{
	glEnable( GL_CULL_FACE );
	glEnable( GL_DEPTH_TEST );
	glDepthFunc( GL_LESS );

	glEnable( GL_LIGHT0 );
}

void CRenderer::SetViewportSize( int nWidth, int nHeight )
{
	m_nWidth = nWidth;
	m_nHeight = nHeight;
	glViewport( 0, 0, nWidth, nHeight );
	SetProjection3D( );
}

void CRenderer::GetViewportSize( int &nWidth, int &nHeight )
{
	nWidth = m_nWidth;
	nHeight = m_nHeight;
}

void CRenderer::SetProjection3D( )
{
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();	
	gluPerspective( 56, (GLdouble)m_nWidth/(GLdouble)m_nHeight, 5.0f, 30000.0f );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();	
}

void CRenderer::SetProjection2D( )
{
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();	
	gluOrtho2D( 0.0, (GLdouble)m_nWidth, 0.0, (GLdouble)m_nHeight );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();	
}

// Setup an orthogonal projection with corners 0,0 and 1,1. Adjust (1,1) corner to maintain aspect ratio.
void CRenderer::SetProjection2D_0_1( )
{
	double X = 1.0;
	double Y = 1.0;
	double Aspect = (GLdouble)m_nWidth / (GLdouble)m_nHeight;
	if ( Aspect > 1.0 )
	{
		X = Aspect;
	} else {
		Y = 1.0 / Aspect;
	}
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluOrtho2D( 0.0, X, 1.0-Y, 1.0 );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();	
}

void CRenderer::SetMatrix(const CMatrix &M)
{
	glLoadMatrixf( M.mf );
}

void CRenderer::SetColor( float r, float g, float b )
{
	glColor3f( r, g, b );
}

void CRenderer::SetLighting( bool bActive )
{
	m_bLighting = bActive;
}

void CRenderer::SetTexturing( bool bActive )
{
	m_bTextured = bActive;
}

void CRenderer::SetLineMode( bool bActive )
{
	if ( bActive ) {
		glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
	} else {
		glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
	}
}

void CRenderer::SetPointSize( float Size )
{
	glPointSize( Size );
}

//
// Drawing Primitives
//
void CRenderer::DrawLine( const CVec3 &P1, const CVec3 &P2 )
{
	glBegin( GL_LINES );
	glVertex3f( P1.x, P1.y, P1.z );
	glVertex3f( P2.x, P2.y, P2.z );
	glEnd();
}

//
// Draw a mesh with openGL
//
void CRenderer::DrawMesh( CVec3 *pVertices, unsigned int *pTris, int nbTris )
{
	glEnableClientState( GL_VERTEX_ARRAY );
	glVertexPointer( 3, GL_FLOAT, sizeof(CVec3), pVertices );

	glDrawElements( GL_TRIANGLES, nbTris * 3, GL_UNSIGNED_INT, pTris );

	glDisableClientState( GL_VERTEX_ARRAY );
}

//
// Draw a textured mesh. 
// Currently there's just one texture loaded, I'll have to write a texture manager eventually
//
void CRenderer::DrawMeshTextured( CVec3 *pVertices, CVec3 *pNormals, unsigned int *pTris, float *pUVs, int nbTris )
{
	if ( m_bLighting ) 
		glEnable( GL_LIGHTING );

	if ( m_bTextured )
		ActivateTexture();

	glEnableClientState( GL_TEXTURE_COORD_ARRAY );
	glTexCoordPointer( 2, GL_FLOAT, 2 * sizeof(GL_FLOAT), pUVs );

	glEnableClientState( GL_NORMAL_ARRAY );
	glNormalPointer( GL_FLOAT, sizeof(CVec3), pNormals );

	DrawMesh( pVertices, pTris, nbTris );

	glDisableClientState( GL_NORMAL_ARRAY );
	glDisableClientState( GL_TEXTURE_COORD_ARRAY );

	glDisable( GL_TEXTURE_2D );
	glDisable( GL_LIGHTING );
}

//
// Draw a wireframe triangle that will end up in front of a solid triangle
//
void CRenderer::HighlightTri( const CVec3 &P1, const CVec3 &P2, const CVec3 &P3 )
{
	glDepthRange( 0.0f, .9999f); // Offset slightly closer to prevent depth fighting

	DrawLine( P1, P2 );
	DrawLine( P2, P3 );
	DrawLine( P3, P1 );

	glDepthRange( 0.0f, 1.0f);
}

//
// Draw a 2D Triangle from UV coordinates
//
void CRenderer::DrawTri2D( const CVec2 &P1, const CVec2 &P2, const CVec2 &P3 )
{
	glBegin( GL_LINES );
	glVertex2f( P1.x, P1.y );
	glVertex2f( P2.x, P2.y );
	glVertex2f( P3.x, P3.y );
	glVertex2f( P1.x, P1.y );
	glEnd();
}

//
// Draw a 2D Rectangle with origin at (StartX, StartY) and size (Width, Height) and UV corners at (0,0) and (1,1)
//
void CRenderer::DrawTexturedRect( float StartX, float StartY, float Width, float Height )
{
	ActivateTexture();

	glBegin( GL_QUADS );
	glTexCoord2f( 0.0f, 0.0f );
	glVertex2f( StartX, StartY );
	glTexCoord2f( 1.0f, 0.0f );
	glVertex2f( StartX + Width, StartY );
	glTexCoord2f( 1.0f, 1.0f );
	glVertex2f( StartX + Width, StartY + Height );
	glTexCoord2f( 0.0f, 1.0f );
	glVertex2f( StartX, StartY + Height );
	glEnd();

	glDisable( GL_TEXTURE_2D );
}

//
// Render an X at location Vert, used to draw attention to a vert
//
void CRenderer::HighlightVert( const CVec3& Vert )
{
	glColor3f( 1.0f, 1.0f, 1.0f );
	DrawLine( Vert -CVec3(-2,2,2), Vert + CVec3(-2,2,2) );
	DrawLine( Vert -CVec3(2,-2,2), Vert + CVec3(2,-2,2) );
	DrawLine( Vert -CVec3(2,2,-2), Vert + CVec3(2,2,-2) );	
}

void CRenderer::DrawPoint( const CVec3& P )
{
	glDepthRange( 0.0f, .9999f); // Offset slightly closer to prevent depth fighting

	glBegin( GL_POINTS );
	glVertex3f( P.x, P.y, P.z );
	glEnd();

	glDepthRange( 0.0f, 1.0f); 
}

void CRenderer::WireRect( float x1, float y1, float x2, float y2 )
{
	DrawLine( CVec3(x1,y1,0), CVec3(x2,y1,0) );
	DrawLine( CVec3(x2,y1,0), CVec3(x2,y2,0) );
	DrawLine( CVec3(x1,y2,0), CVec3(x2,y2,0) );
	DrawLine( CVec3(x1,y1,0), CVec3(x1,y2,0) );
}

//
// Push or pop viewing matrices
//
void CRenderer::PushMatrices()
{
	glMatrixMode( GL_PROJECTION );
	glPushMatrix();
	glMatrixMode( GL_MODELVIEW );
	glPushMatrix();
}

void CRenderer::PopMatrices()
{
	glMatrixMode( GL_PROJECTION );
	glPopMatrix();
	glMatrixMode( GL_MODELVIEW );
	glPopMatrix();
}
// 
// Draw the rectangle used for selection
// 
void CRenderer::DrawSelectionBox( int X1, int Y1, int X2, int Y2 )
{
	Y1 = m_nHeight-Y1;
	Y2 = m_nHeight-Y2;
	
	PushMatrices();
	SetProjection2D();
	
	// Draw stippled rectangle with OpenGL
	glLineStipple( 2, 0xAAAA );
	glEnable( GL_LINE_STIPPLE );
	WireRect( (float)X1, (float)Y1, (float)X2, (float)Y2 );
	glLineStipple( 1, 0xFFFF );
	glDisable( GL_LINE_STIPPLE );

	PopMatrices();
}

bool CRenderer::CheckError( )
{
	return (glGetError()!=0);
}

//
// TEXTURES
//
// Note: These texture functions are not general purpose, I'm starting simple
//
void CRenderer::NewTexture( int Width, int Height, unsigned char* pImage )
{
	// fixme: should release m_BindTexId if needed
	if ( m_nbTextures == 0 )
		{
		glGenTextures( 1, &m_BindTexId );
		m_nbTextures++;
		}

	glBindTexture( GL_TEXTURE_2D, m_BindTexId );

	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR); 
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR ); 
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	//gluBuild2DMipmaps( GL_TEXTURE_2D, GL_RGB, Width, Height, GL_RGB, GL_UNSIGNED_BYTE, pImage);
	glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB, Width, Height, 0, GL_RGB, GL_UNSIGNED_BYTE, pImage );
}

void CRenderer::ActivateTexture( )
{
	if ( m_nbTextures == 0 ) return;

	glEnable( GL_TEXTURE_2D );
	glBindTexture( GL_TEXTURE_2D, m_BindTexId );	
}

//
// Texture updated functions
//
const int MAX_BRUSH_BYTES = 16384;

void CRenderer::UpdateTexture( const CVec2& Coord, const SColor& Color, float BrushSize )
{
	if ( m_nbTextures == 0 ) return;

	const int Width = (int)BrushSize;
	const int Height = (int)BrushSize;
	const int Bpp = 3; // Bytes per pixel
	const int Size = Width*Height;

	if ( Size * Bpp > MAX_BRUSH_BYTES ) return;
	unsigned char Image[ MAX_BRUSH_BYTES ];

	for ( int i = 0; i < Size*Bpp; i+=Bpp )
	{
		Image[ i ] = unsigned char(Color.r * 255.0f);
		Image[ i+1 ] = unsigned char(Color.g * 255.0f);
		Image[ i+2 ] = unsigned char(Color.b * 255.0f);
	}

	int X = int(Coord.x * 256) - Width/2;
	int Y = int(Coord.y * 256) - Height/2;

	glPixelStorei( GL_UNPACK_ALIGNMENT, 1 ); // Image rows are tightly packed, OpenGL defaults to aligned rows

	// Fixme: Handle partially off edge with smaller brush
	if ( X >= 0 && X < 256-Width && Y >= 0 && Y < 256-Height ) 
	{
		glBindTexture( GL_TEXTURE_2D, m_BindTexId );
		glTexSubImage2D( GL_TEXTURE_2D, 0, X, Y, Width, Height, GL_RGB, GL_UNSIGNED_BYTE, Image );
	}
}