//
// Texture Class
//
// For Loading TGA format textures
// It doesn't load all valid TGAs, although support for that could be easily added if necessary
//
#define _CRT_SECURE_NO_DEPRECATE // turn off VS2005 warning
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <math.h>
#include <stdio.h>
#include "vec3.h"
#include "texture.h"
#include "datatransfer.h"
#include "renderInterface.h"

CTexture::CTexture()
{
	m_pImage = 0;
	m_bLoaded = false;
}

CTexture::~CTexture()
{
	Unload();
}

void CTexture::Unload()
{
	delete[] m_pImage;
	m_pImage = 0;

	m_bLoaded = false;
}


//
// TGA Header Structure
//
struct TGA_Header
{
	// This is what all those useless header bytes are. I only plan on using width, height, and bpp.
    char  indentsize;
    char  colormaptype, imagetype;
    short colormapstart, colormaplength;
    char  colormapbits;
    short xOrigin, yOrigin;
    short width, height;
    char  bpp, descriptor;
    
	// We don't just fread the entire header to avoid needing to set struct packing alignment to 1
    void DataIO( CDataTransfer &IO ) {
		IO << indentsize << colormaptype << imagetype << colormapstart << colormaplength << colormapbits;
		IO << xOrigin << yOrigin << width << height << bpp << descriptor;
		}
};

//
// Load a TGA file (sFilename) from the disk.
//
bool CTexture::LoadTGA( const char* Filename )
{
	FILE *pFile = fopen( Filename, "rb" );
	if ( pFile ) 
	{
		// Unload this texture if it's already loaded
		if ( m_bLoaded ) 
		{
			Unload();
		}

		// Load the Header
		TGA_Header Header;
		CDataTransfer LoadFile = CDataTransfer( pFile, CDataTransfer::FILE_LOAD );
		Header.DataIO( LoadFile );

		// Set Member Data from the Header
		m_Width  = Header.width;
		m_Height = Header.height;
		m_bpp	 = Header.bpp;

		if ( m_Width < 0 || m_Height < 0 || m_bpp != 24 ) {
			// We're only loading valid 24-bit tga files, although we could load any valid tga file if we wanted
			fclose( pFile );
			return false;
		}

		// Load the Image Data
		int nbPixels = m_Width * m_Height;
		int BytesPerPixel = m_bpp / 8;

		m_pImage = new unsigned char[ nbPixels * BytesPerPixel + 12 ];

		LoadFile.Data( m_pImage, nbPixels * BytesPerPixel );

		fclose( pFile );

		m_bLoaded = true;
		SwapChannelsRB();
		return true;
	}

	return false;
}

//
// Save a .TGA format image to Filename
//
bool CTexture::SaveTGA( const char* Filename )
{
	FILE *pFile = fopen( Filename, "wb" );
	if ( pFile ) 
	{
		// Create Header
		TGA_Header Header;
		memset( &Header, 0, sizeof(Header) );
		Header.bpp = m_bpp;
		Header.width = m_Width;
		Header.height = m_Height;
		Header.imagetype = 2;

		// Write Data
		CDataTransfer SaveFile = CDataTransfer( pFile, CDataTransfer::FILE_SAVE );

		Header.DataIO( SaveFile );

		int nbPixels = m_Width * m_Height;
		int BytesPerPixel = m_bpp / 8;

		SwapChannelsRB(); // The texture is in memory as RGB, tga uses BGR
		SaveFile.Data( m_pImage, nbPixels * BytesPerPixel );
		SwapChannelsRB();

		fclose( pFile );
		return true;
	}
	return false;
}

// Swap the blue and red channels
void CTexture::SwapChannelsRB( )
{
	int BPP = m_bpp/8; // bytes from bits
	int size = m_Width * m_Height * BPP;
	for ( int i = 0; i < size; i+=BPP )
	{
		unsigned char temp = m_pImage[i];
		m_pImage[i] = m_pImage[i+2];
		m_pImage[i+2] = temp;
	}
}

//
// Add the texture to the renderer
//
void CTexture::BindTexture( CRenderer *pRender )
{
	if ( m_bLoaded )
		pRender->NewTexture( m_Width, m_Height, m_pImage );

		// could free the texture data now to prevent duplication, since openGL has it stored
}

//
// returns the r,g,b values for a location on the texture
//
bool CTexture::GetColor( CVec2 Coord, SColor& Color )
{
	int X = int(Coord.x * m_Width);
	int Y = int(Coord.y * m_Height);

	if ( X < 0 || Y < 0 ) return false;
	if ( X >= m_Width || Y >= m_Height ) return false;

	int BPP = m_bpp/8; 
	int Index = (X + Y * m_Width) * BPP;
	Color.SetUChar( m_pImage[ Index ], m_pImage[ Index+1 ], m_pImage[ Index+2 ] );

	return true;
}