//
// A Mesh Paint demonstration program
// http://www.3dkingdoms.com/weekly
//
// By Jonathan Kreuzer
// Copyright 2006
//
// All of the windows specific stuff is in this file, plus the RenderFrame function
// 
//
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <gl/gl.h>
#include <gl/glu.h>
#include <math.h>
#include "resource.h"

#include "vec3.h"
#include "matrix.h"
#include "renderInterface.h"
#include "mesh.h"
#include "texture.h"
#include "paint.h"
#include "viewport.h"

// Globals
CView g_View;

CRenderer *g_pRenderer = NULL;
CMesh g_Mesh;
CTexture g_Texture;
CSelection g_Select;
CTexPaint g_Paint;

bool g_bWireframe = false;
bool g_bLighting = true;
bool g_bNormals = false;
bool g_bTextured = true;
bool g_bVertices = false;

// Maybe there will be more modes later
enum eViewMode{ VM_MODEL, VM_TEXTURE };
eViewMode g_ViewMode = VM_MODEL;

// Functions
ATOM				MyRegisterClass(HINSTANCE hInstance);
LRESULT CALLBACK	WndProc(HWND, UINT, WPARAM, LPARAM);
void RenderFrame( class CRenderer *pRender );
void MenuUpdateChecks( HWND hWnd );

// Main
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPTSTR lpCmdLine, int nCmdShow)
{
	const int WinWidth = 600;
	const int WinHeight = 600;

	MyRegisterClass(hInstance);
	
	HWND hWnd = CreateWindow( "testappclass", "Sample App: stage 4 ( Paint on Mesh )", WS_OVERLAPPEDWINDOW,
				 CW_USEDEFAULT, 0, WinWidth, WinHeight, NULL, NULL, hInstance, NULL);

    if (!hWnd) return FALSE; 

	// Initialize Renderer
	CRenderer *pRender = new CRenderer;
	g_pRenderer = pRender;
	pRender->InitRenderWindow( hWnd, true, true );
	if ( !pRender->IsInitialized() ) return 0;

	// Show the Window
	MoveWindow( hWnd, 20, 20, WinWidth, WinHeight, TRUE );
    ShowWindow( hWnd, nCmdShow );
    UpdateWindow( hWnd );
	MenuUpdateChecks( hWnd );
	
	// Load the Mesh	
	g_Mesh.Load( "data/meshhead2.dat" );

	// Load a Texture
	g_Texture.LoadTGA( "data/headdiff.tga" );
	g_Texture.BindTexture( pRender );

	// Main loop
	MSG msg;
	bool bDone = false;
	
	while ( !bDone ) 
	{
		// Render the Frame
		RenderFrame( pRender );	
		
		// Check for and process windows messages each frame
		if ( PeekMessage(&msg, NULL, 0, 0, PM_REMOVE ) )
		{		
			if ( msg.message == WM_QUIT ) {
				bDone = true;
			} else {	
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		}
	}

	// Exit
	pRender->Shutdown( );
	return (int) msg.wParam;
}

//
// Called every frame to render the screen using pRender
//
void RenderFrame( CRenderer *pRender )
{
	pRender->ClearBuffer( );
	
	if ( g_ViewMode == VM_MODEL )
	{
		// Rotate the Viewport
		pRender->SetMatrix( g_View.GetMatrix() );
		
		// Render everything
		pRender->SetLighting( g_bLighting );
		pRender->SetTexturing( g_bTextured );

		g_Mesh.Render( pRender, g_bWireframe );
		if ( g_bNormals ) g_Mesh.RenderNormals( pRender, 15.0f );
		if ( g_bVertices ) g_Mesh.RenderVertices( pRender );
		
		//g_Select.Render( pRender );	
	} 

	if ( g_ViewMode == VM_TEXTURE )
	{
		pRender->SetProjection2D_0_1();
		g_Mesh.RenderUVs( pRender );
		pRender->SetProjection3D();
	}

	// Show the back buffer
	pRender->BufferSwap( );	

	// Note: Right now this is just to make sure there are no errors.
	// The renderer should return a specific error string in case an error does occur.
	static bool bCheckError = true;
	if ( bCheckError && pRender->CheckError() ) 
	{
		bCheckError = false;
		MessageBox( GetActiveWindow(), "OpenGL Error.", "Errors", MB_OK | MB_ICONERROR );
	}
}

//
// Update the Menu Item checks based on our variables
//
void MenuUpdateChecks( HWND hWnd )
{
	HMENU hMenu = GetMenu( hWnd );

	CheckMenuItem( hMenu, ID_VIEW_WIREFRAME, ( g_bWireframe ) ? MF_CHECKED : MF_UNCHECKED );
	CheckMenuItem( hMenu, ID_VIEW_LIGHTING, ( g_bLighting ) ? MF_CHECKED : MF_UNCHECKED );
	CheckMenuItem( hMenu, ID_VIEW_NORMALS, ( g_bNormals ) ? MF_CHECKED : MF_UNCHECKED );
	CheckMenuItem( hMenu, ID_VIEW_VERTICES, ( g_bVertices ) ? MF_CHECKED : MF_UNCHECKED );
	CheckMenuItem( hMenu, ID_VIEW_TEXTURED, ( g_bTextured ) ? MF_CHECKED : MF_UNCHECKED );

	if ( g_ViewMode == VM_TEXTURE )  
		CheckMenuItem( hMenu, ID_VIEW_TEXTUREMODE, MF_CHECKED );
	else
		CheckMenuItem( hMenu, ID_VIEW_TEXTUREMODE, MF_UNCHECKED );
}

//
// Get Texture Coordinate in ( Coord ) from mouse position ( X, Y )
// returns false if the mouse position is not on the texture
//
bool GetTexCoord( int X, int Y, CVec2& Coord )
{
	if ( X > 32000 ) X = 0; // Windows negative window coord values get big
	if ( Y > 32000 ) Y = 0;

	if ( g_ViewMode == VM_TEXTURE ) 
	{
		int Width = 0, Height = 0;
		g_pRenderer->GetViewportSize( Width, Height );
		int Size = (Width < Height) ? Width : Height;

		Coord.x = (float)X / (float)Size;
		Coord.y = 1.0f - (float)Y / (float)Size;
		return true;
	}

	if ( g_ViewMode == VM_MODEL ) 
	{
		CVec3 P1, P2;
		g_Select.GetLine( P1, P2, X, Y );
		return g_Mesh.GetTexCoordsFromLine( P1, P2, Coord );
	}

	return false;
}

//
// Menu command processing
//
int MenuCommand( HWND hWnd, int wmId, int wmEvent )
{
switch (wmId)
	{
	case IDM_EXIT:
		DestroyWindow(hWnd);
	return TRUE;

	case ID_VIEW_WIREFRAME:
		g_bWireframe = !g_bWireframe;
		MenuUpdateChecks(hWnd);
	return TRUE;

	case ID_VIEW_LIGHTING:
		g_bLighting = !g_bLighting;
		MenuUpdateChecks(hWnd);
	return TRUE;

	case ID_VIEW_NORMALS:
		g_bNormals = !g_bNormals;
		MenuUpdateChecks(hWnd);
	return TRUE;

	case ID_VIEW_TEXTURED:
		g_bTextured = !g_bTextured;
		MenuUpdateChecks(hWnd);
	return TRUE;

	case ID_VIEW_VERTICES:
		g_bVertices = !g_bVertices;
		MenuUpdateChecks(hWnd);
	return TRUE;

	case ID_VIEW_TEXTUREMODE:
		if ( g_ViewMode == VM_MODEL ) 
		{
			g_ViewMode = VM_TEXTURE;
		} else {
			g_ViewMode = VM_MODEL;
		}
		MenuUpdateChecks(hWnd);
	return TRUE;

	case ID_RELOAD_TEXTURE:
		g_Texture.BindTexture( g_pRenderer );
	return TRUE;
	}
return FALSE;
}

//
// 
//
LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	PAINTSTRUCT ps;
	HDC hdc;
	POINT Point;
	static bool bRight = false, bMiddle = false, bLeft = false;
	static POINT StartPoint = {0,0};
	static CVec2 Coord;
	static SColor Color;
	
	switch (message)
	{
	case WM_COMMAND:
		if ( !MenuCommand( hWnd, LOWORD(wParam), HIWORD(wParam) ) )
			return DefWindowProc(hWnd, message, wParam, lParam);
		break;

	case WM_CHAR:
		if ( LOWORD(wParam) == '[' )
			g_Paint.DecreaseBrushSize();
		if ( LOWORD(wParam) == ']' )
			g_Paint.IncreaseBrushSize();
		break;
		
	case WM_ERASEBKGND:
		// This window is fully drawn by OpenGL, so don't allow any standard window erasing
		// ( ps. I think BKGND is one of the weirdest and hardest to remember abbreviations I've seen )
		return false;
			
	case WM_PAINT:
		hdc = BeginPaint(hWnd, &ps);
		if (g_pRenderer && g_pRenderer->IsInitialized() ) 
			{
			RenderFrame( g_pRenderer );
			}
		EndPaint(hWnd, &ps);
		break;
		
	case WM_DESTROY:
		PostQuitMessage(0);
		break;

	case WM_SIZE:

		g_pRenderer->SetViewportSize( LOWORD(lParam), HIWORD(lParam) );
		g_Select.SetWindowHeight( HIWORD(lParam) ); // keep track of the window height so we can flip the mouse y	
		break;

	case WM_LBUTTONDOWN: 		
	    SetCapture( hWnd ); // So user can drag mouse beyond the window edge without messing things up
	    
		g_Mesh.MarkBackfacing( g_View.GetCamPos() );

		if ( GetTexCoord( LOWORD(lParam), HIWORD(lParam), Coord ) ) 
		{
			if ( GetKeyState( VK_MENU ) < 0 ) {
				// Alt = color-picker
				g_Texture.GetColor( Coord, Color );
				g_Paint.SetBrushColor( Color );
			} else {
				g_Paint.BrushDown( Coord, g_pRenderer );
			}
		}

		bLeft = true;
		break;
		
	case WM_LBUTTONUP:
		ReleaseCapture( );
		g_Mesh.MarkBackfacing( g_View.GetCamPos() );
		/*
		if ( g_ViewMode == VM_MODEL ) {
			g_Select.ButtonUp( &g_Mesh );
		}*/
		g_Paint.BrushUp();
		bLeft = false;
		break;
				
	case WM_RBUTTONDOWN:
		SetCapture( hWnd );
		// store mouse location for the full screen coordinates
		GetCursorPos(&StartPoint); 
		bRight = true;	
		break;
		
	case WM_RBUTTONUP:
		ReleaseCapture( );
		bRight = false;
		break;
	
	case WM_MBUTTONDOWN:
		SetCapture( hWnd );
		GetCursorPos(&StartPoint);
		bMiddle = true;
		break;
		
	case WM_MBUTTONUP:
		ReleaseCapture( );
		bMiddle = false;
		break;
	
	case WM_MOUSEMOVE:
	
		g_Select.MouseMove( LOWORD(lParam), HIWORD(lParam), &g_Mesh );
		GetCursorPos(&Point);

		if ( bLeft && g_Paint.IsBrushDown() )
		{
			if ( GetTexCoord( LOWORD(lParam), HIWORD(lParam), Coord ) ) {
				g_Paint.BrushMove( Coord, g_pRenderer );
			} else {
				g_Paint.BrushUp();
			}
		}
		
		if ( bRight ) {			
			g_View.RotateByDeg( (Point.x - StartPoint.x) / 2.0f, (Point.y - StartPoint.y) / 2.0f );
			StartPoint = Point;
			}
			
		if ( bMiddle ) {
			// Shift zooms, otherwise pan
			if ( GetKeyState( VK_SHIFT ) < 0 ) 
				g_View.DragView( CVec3( 0, 0, -float(Point.y - StartPoint.y ) * 2.0f ) );	
			else
				g_View.DragView( CVec3( float(Point.x - StartPoint.x), -float(Point.y - StartPoint.y ), 0 ) );
				
			StartPoint = Point;
			}
		break;
				
	default:
		return DefWindowProc(hWnd, message, wParam, lParam);
	}
	
	return 0;
}
//
//  
//
ATOM MyRegisterClass(HINSTANCE hInstance)
{
	WNDCLASSEX wcex;
	
	wcex.cbSize = sizeof(WNDCLASSEX);
	wcex.style			= CS_HREDRAW | CS_VREDRAW;
	wcex.lpfnWndProc	= WndProc;
	wcex.cbClsExtra		= 0;
	wcex.cbWndExtra		= 0;
	wcex.hInstance		= hInstance;
	wcex.hIcon			= LoadIcon(hInstance, MAKEINTRESOURCE(IDI_LINEMESHTEST));
	wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground	= (HBRUSH)(COLOR_WINDOW+1);
	wcex.lpszMenuName	= MAKEINTRESOURCE(IDC_LINEMESHTEST);
	wcex.lpszClassName	= "testappclass";
	wcex.hIconSm		= 0;

	return RegisterClassEx(&wcex);
}